#include <stdio.h>
#include <cairo.h>
#include <cairo-pdf.h>
#include <cairo-script.h>
#include <cairo-xml.h>
#include <cairo-svg.h>
#include <gtk/gtk.h>

struct rx {
  cairo_rectangle_t extents;
  cairo_surface_t *recording;
};


static gboolean gtk_expose(GtkWidget *win, GdkEventExpose *event, void *client_data)
{
  struct rx *rx = (struct rx*)client_data;
  cairo_t *cr = gdk_cairo_create(gtk_widget_get_window(win));
  cairo_set_source_surface(cr, rx->recording, -rx->extents.x, -rx->extents.y);
  cairo_paint(cr);
  cairo_destroy(cr);
  return 0;
}


void draw(cairo_surface_t *surface, double x, double y)
{
  cairo_t *cr = cairo_create(surface);
  if ((x != 0.0) && (y != 0.0))
    cairo_translate(cr, x, y);

  cairo_set_line_width(cr, 2.0);
  cairo_save(cr);
  cairo_translate(cr, 10, 10);
  cairo_rectangle(cr, 0, 0, 200, 150);
  cairo_set_source_rgb(cr, 0.5, 0.9, 0.7);
  cairo_fill_preserve(cr);
  cairo_set_source_rgb(cr, 1.0, 0.0, 0.0);
  cairo_stroke(cr);
  cairo_restore(cr);
  cairo_show_page(cr);
  cairo_destroy(cr);
}

void replay(cairo_surface_t *surface, cairo_surface_t *record, double x, double y)
{
  cairo_t *cr = cairo_create(surface);
  cairo_set_source_surface(cr, record, x, y);
  cairo_paint(cr);
  cairo_destroy(cr);
}

int main(int argc, char *argv[])
{
  cairo_rectangle_t r;

  /* --------------------
     Create unbounded recording surface to get ink extents
     -------------------- */

  cairo_surface_t *urecord = cairo_recording_surface_create(CAIRO_CONTENT_COLOR, 0);
  draw(urecord, 0.0, 0.0);

  cairo_recording_surface_ink_extents(urecord, &r.x, &r.y, &r.width, &r.height);
  printf("x=%f y=%f, width=%f height=%f\n", r.x, r.y, r.width, r.height);

  /* --------------------
     Create new bounded recording surface so we can replay it to PDF
     -------------------- */

  cairo_surface_t *record = cairo_recording_surface_create(CAIRO_CONTENT_COLOR, &r);
  draw(record, 0.0, 0.0);

  cairo_surface_t *pdf = cairo_pdf_surface_create("c-record-test.pdf", r.width, r.height);
  draw(pdf, -r.x, -r.y);
  cairo_surface_destroy(pdf);

  pdf = cairo_pdf_surface_create("c-record-test-record-zero.pdf", r.width, r.height);
  replay(pdf, record, 0.0, 0.0);
  cairo_surface_destroy(pdf);

  pdf = cairo_pdf_surface_create("c-record-test-record-offset.pdf", r.width, r.height);
  replay(pdf, record, r.x, r.y);
  cairo_surface_destroy(pdf);

  /* --------------------
     Script direct
     -------------------- */

  cairo_device_t *device = cairo_script_create("c-record-test.cs");
  cairo_surface_t *script = cairo_script_surface_create(device, CAIRO_CONTENT_COLOR, r.width, r.height);
  draw(script, -r.x, -r.y);
  cairo_surface_destroy(script);
  cairo_device_destroy(device);


  /* --------------------
     Script via record playback
     -------------------- */

  device = cairo_script_create("c-record-test-record-zero.cs");
  script = cairo_script_surface_create(device, CAIRO_CONTENT_COLOR, r.width, r.height);
  replay(script, record, 0.0, 0.0);
  cairo_surface_destroy(script);
  cairo_device_destroy(device);

  /* --------------------
     Script via record playback
     -------------------- */

  device = cairo_script_create("c-record-test-record-offset.cs");
  script = cairo_script_surface_create(device, CAIRO_CONTENT_COLOR, r.width, r.height);
  replay(script, record, r.x, r.y);
  cairo_surface_destroy(script);
  cairo_device_destroy(device);

  /* --------------------
     Script via record direct
     -------------------- */

  device = cairo_script_create("c-record-test-record-d.cs");
  cairo_script_from_recording_surface(device, record);
  cairo_device_destroy(device);

  /* --------------------
     xml via record playback
     -------------------- */

  device = cairo_xml_create("c-record-test-record-offset.xml");
  script = cairo_xml_surface_create(device, CAIRO_CONTENT_COLOR, r.width, r.height);
  replay(script, record, r.x, r.y);
  cairo_surface_destroy(script);
  cairo_device_destroy(device);


  /* --------------------
     SVG via record playback
     -------------------- */

  script = cairo_svg_surface_create("c-record-test-record-offset.svg", r.width, r.height);
  replay(script, record, r.x, r.y);
  cairo_surface_destroy(script);
  cairo_device_destroy(device);


  /* --------------------
     xlib via gtk3
     -------------------- */

  struct rx rx;
  rx.extents = r;
  rx.recording = record;

  gtk_init(&argc, &argv);
  GtkWidget *window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  GtkWidget *drawing = gtk_drawing_area_new();
  gtk_container_add(GTK_CONTAINER(window), drawing);

  g_signal_connect(G_OBJECT(window), "destroy", G_CALLBACK(gtk_main_quit), NULL);
  g_signal_connect(G_OBJECT(drawing), "draw", G_CALLBACK(gtk_expose), &rx);
  gtk_widget_set_size_request(GTK_WIDGET(drawing), r.width, r.height);
  gtk_widget_show_all(window);
  gtk_main();
}
